using System;
using System.Configuration;
using System.Security.Principal;
using System.Web.Services;
using System.Xml.Serialization;
using System.IO;
using Timers = System.Timers;
using Microsoft.Exchange.WebServices.Data;
using Microsoft.TeamFoundation.Client;
using PushSubscriber = MsdnMag.Exchange.PushNotificationSubscriber;
using System.Web.Services.Protocols;
using System.Linq;
using Threading = System.Threading;
using System.Threading;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using System.Web;

namespace MsdnMag.Exchange.PushNotificationClient
{
    [WebService(Namespace = "http://www.contoso.com/Email2Workitem/2011")]
    public class PushNotificationClient : WebService, INotificationServiceBinding
    {
        #region Constants

        private const string ERROR_EVENT_TEMPLATE = "{0} {1} - Error text: {2}";
        private const string NOTIFICATION_EVENT_TEMPLATE = "{0} {1} - Event name: {2} [{3}]";
        private const string UNSUBSCRIBE_EVENT_TEMPLATE = "{0} {1} - Event name: Unsubscribe";
        private const string CHECKSTATUS_EVENT_TEMPLATE = "{0} {1} - Event name: Check Status";
        private const string SUBSCRIPTION_TIMEOUT_TEMPLATE = "{0} {1} - Event name: Set Subscription Timeout: {2} milliseconds";
        private const string SUBSCRIPTION_RENEWAL_TEMPLATE = "{0} {1} - Event name: Subscription Renewed after {2} minutes of inactivity. New Subscription Id: {3}";
        private const string EMPTY_URL_ERROR_MESSAGE = "The URL was empty.";
        private const string LOCALHOST_HOST_ERROR_MESSAGE = "The host name specified in the URL is 'localhost'. A real host name is required for a subscription.\r\n" +
            "Suggested URL: {0}.";
        private const string EMPTY_SUBSCRIPTIONID_ERROR_MESSAGE = "SubscriptionId was empty.";
        private const string SOAP_EXCEPTION_TEMPLATE = "Execution failed.\r\nMessage: {0}";
        private const string NEW_EMAIL_SKIPPED_TEMPLATE = "New email skipped since it was not the first email in conversation (sender: {0} - ConversationTopic ({1}): {2})";
        private const string NEW_EMAIL_FROM_EXISTING_THREAD_TEMPLATE = "Creating new email from existing an thread since the first email in conversation has no work item (sender: {0} - ConversationTopic ({1}): {2})";
        private const string ALREADY_ASSIGNED_EMAIL_SKIPPED_TEMPLATE = "Email skipped since it has already been assigned (sender: {0} - ConversationTopic ({1}): {2})";
        private const string WORKITEM_BODY_TEMPLATE =
            "The following request was created from a new email message.<br/>\r\n<br/>\r\n" +
            "<hr/>\r\n" +
            "<table border='1' cellspacing='2' cellpadding='2' width='100%'>" +
            "<tr><td style='background-color:{8}'>\r\n" +
            "<font color=green><b>Sender:</b></font> {0} ({1})<br/>\r\n" +
            "<font color=green><b>Sent:</b></font> {2}<br/>\r\n" +
            "<font color=green><b>Subject:</b></font> {3}\r\n" +
            "</td></tr><tr><td>\r\n" +
            "{4}" +
            "</td></tr></table>" +
            "<br/>\r\n<br/>\r\n" +
            "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~<br/>\r\n" +
            "Conversation Topic: {5}<br/>\r\n" +
            "Conversation Index ({6}): {7}";
        private const string WORKITEM_PARTIAL_BODY_TEMPLATE =
            "<table border='1' cellspacing='2' cellpadding='2' width='100%'>" +
            "<tr><td style='background-color:{4}'>\r\n" +
            "<font color=green><b>Sender:</b></font> {0} ({1})<br/>\r\n" +
            "<font color=green><b>Sent:</b></font> {2}\r\n" +
            "</td></tr><tr><td>\r\n" +
            "{3}" +
            "</td></tr></table>";
        private const string TFS_APPLICATIONNAME_TEMPLATE = "MsdnMagSubscribeSvc [{0}]";
        private const string WORKITEM_FIELDS_TEMPLATE = "MsdnMag.SupportTicket.UniqueIdentifier={0};MsdnMag.SupportTicket.RequesterName={1};" +
            "MsdnMag.SupportTicket.ConversationTopic={2};MsdnMag.SupportTicket.ConversationIndex={3};System.Description={4};" +
            "MsdnMag.SupportTicket.LatestConversationIndex={5};MsdnMag.SupportTicket.RequestSupportTeam={6}";
        private const string WORKITEM_DESCRIPTION_TEMPLATE = "A request work item was created from a new email message." +
            "\r\nPlease navigate to the History tab to view the email details.";
        private const string UPDATE_WORKITEM_FIELDS_TEMPLATE = "MsdnMag.SupportTicket.LatestConversationIndex={0}";
        private const string NEW_WORKITEM_CREATED_TEMPLATE = "New email written to work item {0} (sender: {1})";
        private const string WORKITEM_UPDATED_TEMPLATE = "Work item updated. Id: {0}. New Conversation Index: {1}.";
        private const string EMAIL_TO_WORKITEM_ERROR_TEMPLATE = "Error (email-2-workitem): {0}\r\n--> StackTrace:\r\n{1}";
        private const string SUBSCRIPTION_RENEWAL_ERROR_TEMPLATE = "Error (subscription renewal): {0}\r\n--> StackTrace:\r\n{1}";
        private const string TIMER_ERROR_TEMPLATE = "Error (PollingTimer): {0}\r\n--> StackTrace:\r\n{1}";
        private const string EMAIL_ALREADY_CONVERTED_TO_WORKITEM_TEMPLATE = "Found email already converted to work item. Work Item: {0}. Email Internet Message Id: {1}.";
        private const string SUBJECT_DOES_NOT_MATCH_PREFIX_TEMPLATE = "Email skipped since subject does not match defined prefix. Prefix='{0}'. Subject='{1}'.";
        private const string UPDATEWORKITEM_ERROR_TEMPLATE = "Error (update work item): {0}\r\n--> StackTrace:\r\n{1}";
        private const string NOTIFICATION_EVENTS_LOG_TEMPLATE = "{0}SubscriberEventLog.txt";
        private const string NOTIFICATION_RESPONSE_MESSAGES_LOG_TEMPLATE = "{0}{1}-{2}.xml";
        private const string STATUS_TEMPLATE = "LastStatusEventDateTime: {0}." +
            " PollingIntervalInMilliseconds: {1}." +
            " Status: {2}.";
        private const string QUERY_BY_CONVERSATION_INDEX_WIQL =
            "SELECT [System.Id] FROM WorkItems WHERE [System.TeamProject] = '{0}'  AND  [System.WorkItemType] = 'Support Request' AND [MsdnMag.SupportTicket.ConversationIndex] = '{1}' ORDER BY [System.Id] desc";
        private const string QUERY_BY_UNIQUE_IDENTIFIER_WIQL =
            "SELECT [System.Id] FROM WorkItems WHERE [System.TeamProject] = '{0}'  AND  [System.WorkItemType] = 'Support Request' AND [MsdnMag.SupportTicket.UniqueIdentifier] = '{1}' ORDER BY [System.Id] desc";

        #endregion

        #region Private Fields

        private static object _syncLock = new object();
        private static object _pollingTimerSyncLock = new object();

        // Set the working timer for polling
        private static Threading.Timer _pollingTimer;
        private static DateTime _lastStatusEventDateTime = DateTime.MinValue;

        // Unsubscribe timer
        private static Timers.Timer _timer;
        // Unsubscribe flag
        private static bool _unsubscribe;

        // Static initialization (lazy initialization until first access)
        private static string _baseLogFolder = BaseLogFolder;

        private static int _pollingIntervalInMilliseconds = Convert.ToInt32(ConfigurationManager.AppSettings["AutoSubscribePollingIntervalInMinutes"]) * 60 * 1000; // Every 1 minute.

        #endregion

        #region Properties

        private static string BaseLogFolder
        {
            get
            {
                string baseLogFolder = (string.IsNullOrWhiteSpace(ConfigurationManager.AppSettings["BaseLogsFolder"]) ?
                    string.Empty : string.Format("{0}{1}", ConfigurationManager.AppSettings["BaseLogsFolder"],
                    (ConfigurationManager.AppSettings["BaseLogsFolder"].EndsWith("\\") ? string.Empty : "\\")));

                if (baseLogFolder.Length == 0)
                {
                    return baseLogFolder;
                }

                string absolutePath;
                // If the path is relative, calculate absolute path and return it.
                if (!Path.IsPathRooted(baseLogFolder))
                {
                    string combinedPath = Path.Combine(HttpRuntime.AppDomainAppPath, baseLogFolder);
                    absolutePath = Path.GetFullPath(combinedPath);
                }
                else // path is already absolute.
                {
                    absolutePath = baseLogFolder;
                }

                if (!Directory.Exists(absolutePath))
                {
                    Directory.CreateDirectory(absolutePath);
                }

                return absolutePath;
            }
        }

        private static string TfsClientCacheDirectory
        {
            get
            {
                string tfsClientCacheDirectory = ConfigurationManager.AppSettings["TfsClientCacheDirectory"];

                if (string.IsNullOrWhiteSpace(tfsClientCacheDirectory))
                {
                    return string.Empty;
                }

                // If the path is relative, calculate absolute path and return it.
                if (!Path.IsPathRooted(tfsClientCacheDirectory))
                {
                    string combinedPath = Path.Combine(HttpRuntime.AppDomainAppPath, tfsClientCacheDirectory);
                    return Path.GetFullPath(combinedPath);
                }
                else // path is already absolute.
                {
                    return tfsClientCacheDirectory;
                }
            }
        }

        #endregion

        #region Constructor

        public PushNotificationClient()
        {
            InitializeTimer();
        }

        #endregion

        #region Private Methods

        private void InitializeTimer()
        {
            int autoSubscribeTimeoutThresholdInMinutes = Convert.ToInt32(ConfigurationManager.AppSettings["AutoSubscribeTimeoutThresholdInMinutes"]);

            lock (_pollingTimerSyncLock)
            {
                if (autoSubscribeTimeoutThresholdInMinutes > 0)
                {
                    // Seed the timer with the polling interval specified in config.
                    if (_pollingTimer == null)
                    {
                        TimerCallback timerCallback = OnTimedEvent;
                        _pollingTimer = new Threading.Timer(timerCallback, null, 0, _pollingIntervalInMilliseconds); // Enable the timer.
                    }
                }
                else
                {
                    // Dispose of the timer.
                    if (_pollingTimer != null)
                    {
                        _pollingTimer.Dispose();
                        _pollingTimer = null;
                    }
                }
            }
        }

        #endregion

        #region INotificationServiceBinding Members

        public SendNotificationResultType SendNotification(SendNotificationResponseType sendNotification)
        {
            lock (_pollingTimerSyncLock)
            {
                _lastStatusEventDateTime = DateTime.Now;
            }

            string timestamp = PushSubscriber.PushNotificationSubscriber.GetTimestamp();

            // Create the response to the push notification.
            SendNotificationResultType result = new SendNotificationResultType();

            // Get the push notifications.
            ResponseMessageType[] rmta = sendNotification.ResponseMessages.Items;

            foreach (ResponseMessageType rmt in rmta)
            {
                if (rmt.ResponseCode != ResponseCodeType.NoError)
                {
                    // Log the error text.
                    WriteEventToLog(string.Format(ERROR_EVENT_TEMPLATE,
                        DateTime.Now.ToShortDateString(), DateTime.Now.ToLongTimeString(), rmt.MessageText));

                    // Discontinue the subscription if there is an error.
                    result.SubscriptionStatus = SubscriptionStatusType.Unsubscribe;
                    return result;
                }

                // Cast to the correct response message type.
                SendNotificationResponseMessageType snrmt = rmt as SendNotificationResponseMessageType;

                if (rmt.ResponseCode != ResponseCodeType.NoError
                    || (rmt.ResponseCode == ResponseCodeType.NoError && Convert.ToBoolean(ConfigurationManager.AppSettings["LogStatusEventResponse"].ToLowerInvariant()) == true))
                {
                    // Write the XML content of the push notification.
                    CreateXmlMessageTextFile(snrmt, timestamp);
                }

                // Get the notification.
                NotificationType notification = snrmt.Notification;

                // Get the subscription identifier.
                string subscriptionId = notification.SubscriptionId;

                // Get the previous watermark for the subscription.
                string previousWatermark = notification.PreviousWatermark;

                // Indicate if an event has occured.
                bool eventOccured = false;

                // Handle each notification in the order in which the events occured.
                for (int c = 0; c < notification.Items.Length; c++)
                {
                    // Get the event name.
                    WriteEventToLog(string.Format(NOTIFICATION_EVENT_TEMPLATE,
                        DateTime.Now.ToShortDateString(), DateTime.Now.ToLongTimeString(), notification.ItemsElementName[c].ToString(), subscriptionId));

                    BaseNotificationEventType[] bnet = notification.Items;

                    // Access information about copy or move events.
                    if (bnet[c] is MovedCopiedEventType)
                    {
                        AccessMovedCopiedEvent(bnet[c] as MovedCopiedEventType);
                        eventOccured = true;

                        lock (_syncLock)
                        {
                            // Close the _timer.
                            if (_timer != null)
                                _timer.Close();
                        }
                    }

                    // Access information about create, delete, or new mail events.
                    else if (bnet[c] is BaseObjectChangedEventType)
                    {
                        AccessCreateDeleteNewMailEvent(bnet[c] as BaseObjectChangedEventType);
                        eventOccured = true;

                        lock (_syncLock)
                        {
                            // Close the _timer.
                            if (_timer != null)
                                _timer.Close();
                        }
                    }

                    // Access information about modified events.
                    else if (bnet[c] is ModifiedEventType)
                    {
                        AccessModifiedEvent(bnet[c] as ModifiedEventType);
                        eventOccured = true;

                        lock (_syncLock)
                        {
                            // Close the _timer.
                            if (_timer != null)
                                _timer.Close();
                        }
                    }

                    // No events have occured.
                    else
                    {
                        double unsubscribeOnIdleTimeoutInMilliseconds = Convert.ToDouble(ConfigurationManager.AppSettings["UnsubscribeOnIdleTimeoutInMilliseconds"]);
                        bool isUnsubscribeOnIdle = Convert.ToBoolean(ConfigurationManager.AppSettings["UnsubscribeOnIdle"]);
                        if (!eventOccured && isUnsubscribeOnIdle)
                        {
                            WriteEventToLog(string.Format(SUBSCRIPTION_TIMEOUT_TEMPLATE,
                                DateTime.Now.ToShortDateString(), DateTime.Now.ToLongTimeString(), unsubscribeOnIdleTimeoutInMilliseconds));

                            lock (_syncLock)
                            {
                                // Set _timer to n minutes (web.config). If no events have occured in n minutes, then set _unsubscribe flag.
                                _timer = new Timers.Timer(unsubscribeOnIdleTimeoutInMilliseconds);
                                _timer.Elapsed += new Timers.ElapsedEventHandler(timer_Elapsed);
                                _timer.Enabled = true;
                            }

                        }
                        else
                        {
                            eventOccured = false;
                        }

                    }
                }

                // If there are more events ready to be sent, then continue the subscription.
                if (notification.MoreEvents)
                {
                    result.SubscriptionStatus = SubscriptionStatusType.OK;
                    return result;
                }
            }

            bool unsubscribe = _unsubscribe;
            if (unsubscribe)
            {
                WriteEventToLog(string.Format(UNSUBSCRIBE_EVENT_TEMPLATE,
                    DateTime.Now.ToShortDateString(), DateTime.Now.ToLongTimeString()));

                result.SubscriptionStatus = SubscriptionStatusType.Unsubscribe;
                return result;
            }

            else
            {
                result.SubscriptionStatus = SubscriptionStatusType.OK;
                return result;
            }
        }

        #endregion

        #region Web Methods

        /// <summary>
        /// Checks the status of the web service and makes sure that all configured values are fine.
        /// </summary>
        /// <returns>A string describing the status of the web service.</returns>
        [WebMethod(Description = "Checks the status of the web service and makes sure that all configured values are fine.")]
        public string CheckStatus()
        {
            string lastUpdateDateTime = _lastStatusEventDateTime == DateTime.MinValue ?
                "-" :
                string.Format("{0} {1}", _lastStatusEventDateTime.ToShortDateString(), _lastStatusEventDateTime.ToLongTimeString());
            int seedValueInMilliseconds = 60000; // 1 minute.
            string status = _lastStatusEventDateTime == DateTime.MinValue ?
                "Not subscribed" :
                (_lastStatusEventDateTime.AddMilliseconds(_pollingIntervalInMilliseconds + seedValueInMilliseconds).CompareTo(DateTime.Now) >= 0) ? "Subscribed" : "Not Subscribed";

            string statusMessage = string.Format(
                STATUS_TEMPLATE,
                lastUpdateDateTime,
                _pollingIntervalInMilliseconds,
                status);

            WriteEventToLog(string.Format(CHECKSTATUS_EVENT_TEMPLATE,
                DateTime.Now.ToShortDateString(), DateTime.Now.ToLongTimeString()));

            return statusMessage;
        }

        /// <summary>
        /// Creates a push notification subscription for new email messages for the current web service's App Pool identity.
        /// The subscription is registered against the current web service's URL.
        /// </summary>
        /// <returns>A string representing the Subscription Id for a successful subscription.</returns>
        [WebMethod(Description = "Creates a push notification subscription for new email messages for the current web service's App Pool identity. " +
            "The subscription is registered against the current web service's URL. Do not call this method if the host name in the URL is 'localhost'. " +
            "Make sure it is the actual host name or the DNS name.")]
        public string Subscribe()
        {
            try
            {
                string url = (Context.Request.Url.AbsoluteUri.ToString() ?? string.Empty).Replace("/Subscribe", string.Empty);
                if (string.IsNullOrWhiteSpace(url))
                {
                    throw new Exception(EMPTY_URL_ERROR_MESSAGE);
                }

                if (string.Compare(Context.Request.Url.Host, "localhost", ignoreCase: true) == 0)
                {
                    string hostName = Context.Server.MachineName;
                    string errorMessage = string.Format(
                        LOCALHOST_HOST_ERROR_MESSAGE,
                        url.Replace("://localhost", "://" + hostName));

                    throw new Exception(errorMessage);
                }

                PushSubscriber.PushNotificationSubscriber pushNotificationSubscriber = new PushSubscriber.PushNotificationSubscriber();
                pushNotificationSubscriber.BaseLogFolder = _baseLogFolder;
                pushNotificationSubscriber.EwsServiceUrl = ConfigurationManager.AppSettings["EWSServiceUrl"];
                pushNotificationSubscriber.PushClientWebServiceUrl = url;
                pushNotificationSubscriber.UseCurrentlyLoggedInUserCredentials = true;
                pushNotificationSubscriber.UseAutodiscoveryForSubscribing = Convert.ToBoolean(ConfigurationManager.AppSettings["UseAutodiscoveryForSubscribing"]);
                pushNotificationSubscriber.AutoDiscoveryEmail = ConfigurationManager.AppSettings["AutoDiscoveryEmail"];
                pushNotificationSubscriber.PushNotificationsFrequencyInMinutes = Convert.ToInt32(ConfigurationManager.AppSettings["PushNotificationsFrequencyInMinutes"]);
                pushNotificationSubscriber.SubscribeForPushNotifications();

                string subscriptionId = pushNotificationSubscriber.SubscriptionId;
                if (string.IsNullOrWhiteSpace(subscriptionId))
                {
                    throw new Exception(EMPTY_SUBSCRIPTIONID_ERROR_MESSAGE);
                }

                // Reset the status time to avoid registring multiple redudant subscriptions.
                lock (_pollingTimerSyncLock)
                {
                    _lastStatusEventDateTime = DateTime.Now;
                }

                return subscriptionId;
            }
            catch (Exception ex)
            {
                throw new SoapException(
                    string.Format(SOAP_EXCEPTION_TEMPLATE, ex.Message),
                    SoapException.ClientFaultCode,
                    ex);
            }
        }

        #endregion

        #region Timer Callback

        /// <summary>
        /// Timer Elapsed event callback.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        /// <remarks>
        /// Unsubscribe if the _timer runs out.
        /// </remarks>
        void timer_Elapsed(object sender, Timers.ElapsedEventArgs e)
        {
            lock (_syncLock)
            {
                _unsubscribe = true;

                // Close the _timer.
                if (_timer != null)
                    _timer.Close();
            }
        }

        private void OnTimedEvent(object stateInfo)
        {
            try
            {
                lock (_pollingTimerSyncLock)
                {
                    // Disable the timer to avoid cascading events.
                    _pollingTimer.Change(Timeout.Infinite, _pollingIntervalInMilliseconds);
                }

                int autoSubscribeTimeoutThresholdInMinutes = Convert.ToInt32(ConfigurationManager.AppSettings["AutoSubscribeTimeoutThresholdInMinutes"]);

                // Verify the last time we got a status message was < "AutoSubscribeTimeoutThresholdInMinutes" minutes ago.
                DateTime currentTime = DateTime.Now;
                if (_lastStatusEventDateTime == DateTime.MinValue) // first time or never subscribed, so in that case, we don't want to force a subscription until it is done at least once.
                {
                    return;
                }

                else if (_lastStatusEventDateTime.AddMinutes(autoSubscribeTimeoutThresholdInMinutes) < currentTime) // Last status notification was more than "AutoSubscribeTimeoutThresholdInMinutes" minutes ago...
                {
                    // Renew the subscription.
                    try
                    {
                        string subscriptionId = Subscribe();
                        WriteEventToLog(string.Format(SUBSCRIPTION_RENEWAL_TEMPLATE,
                            DateTime.Now.ToShortDateString(), DateTime.Now.ToLongTimeString(), autoSubscribeTimeoutThresholdInMinutes, subscriptionId));
                    }
                    catch (SoapException ex)
                    {
                        WriteEventToLog(string.Format(SUBSCRIPTION_RENEWAL_ERROR_TEMPLATE, ex.Message, ex.StackTrace));
                    }
                }
            }
            catch (Exception ex)
            {
                WriteEventToLog(string.Format(TIMER_ERROR_TEMPLATE, ex.Message, ex.StackTrace));
            }
            finally
            {
                lock (_pollingTimerSyncLock)
                {
                    _pollingTimer.Change(_pollingIntervalInMilliseconds, _pollingIntervalInMilliseconds); // re-enable the timer.
                }
            }
        }

        #endregion

        #region Static Helper Methods

        /// <summary>
        /// Access event information for moved or copied events.
        /// </summary>
        /// <param name="mcet">The event details for a moved or copied event.</param>
        private static void AccessMovedCopiedEvent(MovedCopiedEventType mcet)
        {
            // Get the watermark for the event.
            string watermark = mcet.Watermark;

            // Get the timestamp for the event.
            DateTime timestamp = mcet.TimeStamp;

            // Get the old parent folder identifier.
            FolderIdType oldParentFolder = mcet.OldParentFolderId;

            // Get the new parent folder identifier.
            FolderIdType newParentFolder = mcet.ParentFolderId;

            // Get either the old folder identifier for a moved/copied folder
            // or the old item identifier for a moved/copid item.
            if (mcet.Item1 is ItemIdType)
            {
                // Get the old item identifier of the item.
                ItemIdType oldItemId = mcet.Item1 as ItemIdType;
            }
            else
            {
                // Get the old folder identifier of the folder.
                FolderIdType oldFolderId = mcet.Item1 as FolderIdType;
            }

            // Get either the new folder identifier for a moved/copied folder
            // or the new item identifier fir a moved/copied item.
            if (mcet.Item is ItemIdType)
            {
                // Get the new item identifier of the item.
                ItemIdType newItemId = mcet.Item as ItemIdType;
            }
            else
            {
                // Get the new folder identifier of the folder.
                FolderIdType newFolderId = mcet.Item as FolderIdType;
            }
        }

        /// <summary>
        /// Access event information for create, delete, or new mail events.
        /// </summary>
        /// <param name="bocet">The event details for a create, delete or new email event.</param>
        private static void AccessCreateDeleteNewMailEvent(BaseObjectChangedEventType bocet)
        {
            // Get the watermark for the event.
            string watermark = bocet.Watermark;

            // Get the timestamp for the event.
            DateTime timestamp = bocet.TimeStamp;

            // Get the parent folder identifier.
            FolderIdType parentFolderId = bocet.ParentFolderId;

            // Get either the folder or item identifier for a 
            // create/delete/new mail event.
            if (bocet.Item is ItemIdType)
            {
                // Get the item identifier.
                ItemIdType itemId = bocet.Item as ItemIdType;

                //
                // Get the message using the ItemIdType.
                //

                // From web.config
                string ewsServiceUrl = ConfigurationManager.AppSettings["EWSServiceUrl"];

                ExchangeService esb = new ExchangeService(ExchangeVersion.Exchange2010);
                esb.Url = new Uri(ewsServiceUrl);
                // From web service's app pool identity.
                esb.UseDefaultCredentials = true;

                EmailMessage message = EmailMessage.Bind(esb, new ItemId(itemId.Id));
                ConvertEmailToRequestItem(message);
            }
            else
            {
                // Get the folder identifier.
                FolderIdType folderId = bocet.Item as FolderIdType;
            }
        }

        /// <summary>
        /// Access event information for modified events.
        /// </summary>
        /// <param name="met">The event details for a modified event.</param>
        private static void AccessModifiedEvent(ModifiedEventType met)
        {
            // Get the watermark for the event.
            string watermark = met.Watermark;

            // Get the timestamp for the event.
            DateTime timestamp = met.TimeStamp;

            // Get the parent folder identifier.
            FolderIdType parentFolderId = met.ParentFolderId;

            // Get either the folder or item identifier for the
            // modified folder or item.
            if (met.Item is FolderIdType)
            {
                // Check if the folder has an unread count.
                if (met.UnreadCountSpecified)
                {
                    int unreadCount = met.UnreadCount;
                }

                // Get the folder identifier.
                FolderIdType folderId = met.Item as FolderIdType;
            }
            else
            {
                // Get the item identifier.
                ItemIdType itemId = met.Item as ItemIdType;
            }
        }

        /// <summary>
        /// Converts a new email message into a Request work item in TFS.
        /// </summary>
        /// <param name="message">An instance of EmailMessage containing the mail message details.</param>
        private static void ConvertEmailToRequestItem(EmailMessage message)
        {
            try
            {
                // TFS customizations.
                TfsConnection.ApplicationName = string.Format(TFS_APPLICATIONNAME_TEMPLATE, WindowsIdentity.GetCurrent().Name);
                TfsConnection.ClientSettingsDirectory = TfsClientCacheDirectory;

                // A conversation index is 22 bytes in length for the first email in the thread
                // An additional 5-bytes are added for each email in the thread.
                bool isFirstEmailInThread = message.ConversationIndex.Length == 22;

                string conversationIndex = Convert.ToBase64String(message.ConversationIndex);
                string latestConversationIndex = conversationIndex;
                string conversationTopic = message.ConversationTopic;
                string internetMessageId = message.InternetMessageId;

                var title = message.Subject;
                var from = message.From;

                // First check if prefix-based filtering is turned on, and if so, if the email subject line matches
                // the defined prefix.
                string subjectPrefix = ConfigurationManager.AppSettings["SubjectPrefix"];
                if (!string.IsNullOrWhiteSpace(subjectPrefix))
                {
                    if (!title.StartsWith(subjectPrefix, StringComparison.InvariantCultureIgnoreCase))
                    {
                        WriteEventToLog(
                            string.Format(SUBJECT_DOES_NOT_MATCH_PREFIX_TEMPLATE,
                            subjectPrefix, title));

                        return;
                    }
                }

                // Check if there is already a work item with the same internal message id.
                // Since this identifier is unique to each message, if we find it already associated with a work
                // item, then this is typically an indication that there are multiple subscriptions associated with
                // the web service (possibly unintentionally since the subscription can be triggered multiple times)
                bool isAlreadyAssigned = IsEmailAlreadyConvertedToWorkItem(internetMessageId);
                if (isAlreadyAssigned)
                {
                    WriteEventToLog(
                        string.Format(ALREADY_ASSIGNED_EMAIL_SKIPPED_TEMPLATE,
                        from, message.ConversationIndex.Length, conversationTopic));

                    return;
                }

                if (!isFirstEmailInThread)
                {
                    // Try and find the base work item so we can append thread to it.
                    byte[] baseConversationIndex = message.ConversationIndex.Subset(0, 22);
                    string baseConversationIndexEncoded = Convert.ToBase64String(baseConversationIndex);

                    int workItemId = GetWorkItemByConversation(baseConversationIndexEncoded);
                    if (workItemId > 0)
                    {
                        // Update the work item with the body section unique to this part of the conversation/thread.
                        // If Exchange cannot figure out the unique body, revert to saving the full body...
                        string partialBody =
                            string.Format(WORKITEM_PARTIAL_BODY_TEMPLATE,
                                from.Name,
                                from.Address,
                                string.Format("{0} {1}", message.DateTimeReceived.ToLongDateString(), message.DateTimeReceived.ToLongTimeString()),
                                message.UniqueBody != null ? message.UniqueBody.Text : message.Body.Text,
                                ConfigurationManager.AppSettings["EmailThreadChildHeaderBGColor"]);

                        UpdateWorkItem(workItemId,
                            conversationIndex, /* Store the latest conversation index */
                            partialBody);

                        return;
                    }
                    else
                    {
                        // Since we can't find an existing base work item to which we can append this thread, we can
                        // create a new base work item from this one.
                        // This is typically the case when there are existing threads that have started before this
                        // system was put in use.

                        if (Convert.ToBoolean(ConfigurationManager.AppSettings["CreateNewWorkItemFromExistingThreadWithNoBase"]))
                        {
                            // Store the extracted base conversation index.
                            conversationIndex = baseConversationIndexEncoded;
                            title = conversationTopic; // this removes any RE: and FW: prefixing the email subject.

                            WriteEventToLog(
                                string.Format(NEW_EMAIL_FROM_EXISTING_THREAD_TEMPLATE,
                                from, message.ConversationIndex.Length, conversationTopic));
                        }
                        else
                        {
                            // Skip and return. Mention it in the log.
                            WriteEventToLog(
                                string.Format(NEW_EMAIL_SKIPPED_TEMPLATE,
                                from, message.ConversationIndex.Length, conversationTopic));

                            return;
                        }
                    }
                }

                var body = string.Format(WORKITEM_BODY_TEMPLATE,
                    from.Name,
                    from.Address,
                    string.Format("{0} {1}", message.DateTimeReceived.ToLongDateString(), message.DateTimeReceived.ToLongTimeString()),
                    title,
                    message.Body.Text,
                    conversationTopic,
                    message.ConversationIndex.Length,
                    conversationIndex,
                    ConfigurationManager.AppSettings["EmailThreadParentHeaderBGColor"]);

                // NOTE: The creation logic utilizes an MSBuild task called CreateNewWorkItem to perform the creation.
                //       Since this code is used outside of any MSBuild scripts, the only thing needed in order
                //       to be able to utilize this common task is to pass it a mock IBuildEngine class.
                //       This was done for two reasons:
                //       1- To utilize an existing task class that is well tested.
                //       2- To demonstrate a way of calling and leveraging existing libraries without having to 
                //          rewrite the code to adapt it to a different hosting environment.
                //       The same was done for the update functionality.

                // Create the work item from the email.
                CreateNewWorkItem newWorkItem = new CreateNewWorkItem();
                newWorkItem.BuildEngine = new BuildEngine(); ;
                newWorkItem.WorkItemTeamProject = ConfigurationManager.AppSettings["WorkItemTeamProject"];
                newWorkItem.TeamFoundationServerUrl = ConfigurationManager.AppSettings["TeamFoundationServerUrl"];
                newWorkItem.AssociatedWorkItems = ConfigurationManager.AppSettings["AssociatedWorkItems"];
                newWorkItem.Title = title;
                newWorkItem.Description = body;
                newWorkItem.WorkItemType = ConfigurationManager.AppSettings["WorkItemType"];
                newWorkItem.WorkItemFieldValues = string.Format(
                    WORKITEM_FIELDS_TEMPLATE,
                    internetMessageId,
                    from.Name,
                    conversationTopic,
                    conversationIndex,
                    WORKITEM_DESCRIPTION_TEMPLATE,
                    latestConversationIndex,
                    ConfigurationManager.AppSettings["RequestSupportTeamName"]);

                bool result = newWorkItem.Execute();
                if (result)
                {
                    WriteEventToLog(string.Format(NEW_WORKITEM_CREATED_TEMPLATE, newWorkItem.WorkItemId, from));
                }
            }
            catch (Exception ex)
            {
                WriteEventToLog(string.Format(EMAIL_TO_WORKITEM_ERROR_TEMPLATE, ex.Message, ex.StackTrace));
            }
        }

        /// <summary>
        /// Determines whether the email message has already been written to a work item.
        /// </summary>
        /// <param name="internetMessageId">The Internet Message Id that uniquely identifies any Exchange email message.</param>
        /// <returns>true if the message has already been written to a work item, false otherwise.</returns>
        public static bool IsEmailAlreadyConvertedToWorkItem(string internetMessageId)
        {
            // Query TFS to get the work item id with the specified Internet Message Id.
            string wiql = string.Format(QUERY_BY_UNIQUE_IDENTIFIER_WIQL,
                ConfigurationManager.AppSettings["WorkItemTeamProject"],
                internetMessageId);

            WorkItemCollection workItems = GetWorkItemByCustomQuery(wiql);

            bool isAlreadyConverted = (workItems != null && workItems.Count > 0);
            if (isAlreadyConverted)
            {
                WriteEventToLog(string.Format(EMAIL_ALREADY_CONVERTED_TO_WORKITEM_TEMPLATE, workItems[0].Id, internetMessageId));
            }

            return isAlreadyConverted;
        }

        /// <summary>
        /// Gets a work item with the base conversation index specified.
        /// </summary>
        /// <param name="baseConversationIndex">The base conversation id (first 22 bytes in a conversation index byte array).</param>
        /// <returns>The id of the work item if found, otherwise 0.</returns>
        private static int GetWorkItemByConversation(string baseConversationIndex)
        {
            // Query TFS to get the work item id corresponding to the base conversation index.
            string wiql = string.Format(QUERY_BY_CONVERSATION_INDEX_WIQL,
                ConfigurationManager.AppSettings["WorkItemTeamProject"],
                baseConversationIndex);

            WorkItemCollection workItems = GetWorkItemByCustomQuery(wiql);

            if (workItems != null && workItems.Count > 0)
            {
                // Pick the first work item in the collection; since the query sorts by Id descending, that would be the most recent work item.
                return workItems[0].Id;
            }
            else
            {
                return 0;
            }
        }

        /// <summary>
        /// Executes a WIQL query and returns the resulting work items collection.
        /// </summary>
        /// <param name="wiql">The WIQL query to be executed.</param>
        /// <returns>The resulting WorkItemCollection instance.</returns>
        public static WorkItemCollection GetWorkItemByCustomQuery(string wiql)
        {
            TfsTeamProjectCollection tfs = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(new Uri(ConfigurationManager.AppSettings["TeamFoundationServerUrl"]));
            WorkItemStore workItemStore = (WorkItemStore)tfs.GetService(typeof(WorkItemStore));
            return workItemStore.Query(wiql);
        }

        /// <summary>
        /// Updates a work item with the part of the email body that is unique to the current email in a threaded email conversation.
        /// </summary>
        /// <param name="workItemId">The id of the work item to be updated.</param>
        /// <param name="conversationIndex">The current value of the conversation id.</param>
        /// <param name="uniqueBodyText">The part of the email body that is unique to the current email.</param>
        private static void UpdateWorkItem(int workItemId, string conversationIndex, string uniqueBodyText)
        {
            try
            {
                // NOTE: The update logic utilizes an MSBuild task called UpdateWorkItem to perform the update.
                //       Since this code is used outside of any MSBuild scripts, the only thing needed in order
                //       to be able to utilize this common task is to pass it a mock IBuildEngine class.
                //       This was done for two reasons:
                //       1- To utilize an existing task class that is well tested.
                //       2- To demonstrate a way of calling and leveraging existing libraries without having to 
                //          rewrite the code to adapt it to a different hosting environment.
                //       The same was done for the creation functionality.

                // Update the work item with the latest conversation index and the text of the email thread
                //       that is unique to this last message.
                UpdateWorkItem existingWorkItem = new UpdateWorkItem();
                existingWorkItem.BuildEngine = new BuildEngine(); ;
                existingWorkItem.WorkItemTeamProject = ConfigurationManager.AppSettings["WorkItemTeamProject"];
                existingWorkItem.TeamFoundationServerUrl = ConfigurationManager.AppSettings["TeamFoundationServerUrl"];
                existingWorkItem.AssociatedWorkItems = ConfigurationManager.AppSettings["AssociatedWorkItems"];
                existingWorkItem.WorkItemFieldValues = string.Format(
                    UPDATE_WORKITEM_FIELDS_TEMPLATE,
                    conversationIndex);
                existingWorkItem.WorkItemId = workItemId;
                existingWorkItem.History = uniqueBodyText;

                bool result = existingWorkItem.Execute();
                if (result)
                {
                    WriteEventToLog(string.Format(WORKITEM_UPDATED_TEMPLATE, workItemId, conversationIndex));
                }
            }
            catch (Exception ex)
            {
                WriteEventToLog(string.Format(UPDATEWORKITEM_ERROR_TEMPLATE, ex.Message, ex.StackTrace));
            }
        }

        /// <summary>
        /// Writes an entry into the events log file.
        /// </summary>
        /// <param name="message">The status message to be logged.</param>
        internal static void WriteEventToLog(string message)
        {
            string filename = string.Format(NOTIFICATION_EVENTS_LOG_TEMPLATE, _baseLogFolder);
            int archivalThresholdInBytes = Convert.ToInt32(ConfigurationManager.AppSettings["ArchivalThresholdInKBytes"]) * 1024;

            // Do we need to create a new file and rename the old one?
            if (File.Exists(filename))
            {
                FileInfo fileInfo = new FileInfo(filename);
                if (fileInfo.Length > archivalThresholdInBytes)
                {
                    // Time to archive.
                    string timestamp = PushSubscriber.PushNotificationSubscriber.GetTimestamp();
                    RetryManager.InvokeWithRetries(
                        () => fileInfo.MoveTo(string.Format("{0}-{1}.txt", filename.Substring(0, filename.Length - 4), timestamp)));

                    // Do we need to cleanup?
                    int numberOfArchivesToKeep = Convert.ToInt32(ConfigurationManager.AppSettings["NumberOfArchivesToKeep"]);
                    string[] files = Directory.GetFiles(Directory.GetParent(filename).FullName, "SubscriberEventLog*", SearchOption.TopDirectoryOnly);

                    if (files.Count() > numberOfArchivesToKeep)
                    {
                        // Sort by newest first
                        var sortedFiles = from file in files
                                          orderby new FileInfo(file).LastWriteTime descending
                                          select file;
                        int counter = 0;
                        foreach (var file in sortedFiles)
                        {
                            if (counter >= numberOfArchivesToKeep)
                            {
                                // Delete old archived file.
                                RetryManager.InvokeWithRetries(() => new FileInfo(file).Delete());
                            }
                            counter++;
                        }
                    }
                }
            }
            using (StreamWriter sw = new StreamWriter(filename, true))
            {
                sw.WriteLine(message);
            }
        }

        /// <summary>
        /// Persists the SendNotificationResponseMessageType message to an XML file.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="timestamp">A timestamp string (formatted datetime value).</param>
        public static void CreateXmlMessageTextFile(SendNotificationResponseMessageType response, string timestamp)
        {
            // Create a text file with the notification XML.
            using (StreamWriter myRespWriter = new StreamWriter(string.Format(NOTIFICATION_RESPONSE_MESSAGES_LOG_TEMPLATE, _baseLogFolder, response.GetType().Name, timestamp)))
            {
                XmlSerializer mySerializer = new XmlSerializer(response.GetType());
                mySerializer.Serialize(myRespWriter, response);
            }
        }

        #endregion
    }
}
